/*  vim600: sts=4
 *  Authors: Valery Kotchiev <valery@diku.dk>
 *  A simple X volume control with OSD drawing.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Xresource.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>   /* This is needed on RedHat systems */
#include <unistd.h>
#include <signal.h>

#include <fcntl.h> /* For mixer open() */
#include <sys/ioctl.h>

#ifdef FREEBSD
#include <machine/soundcard.h>
#endif
#ifdef LINUX
#include <linux/soundcard.h>
#endif


/****************************************************************
 * 
 * Global variables
 *
 ***************************************************************/
Display *display;
Window mainwindow;
GC gcon;
Pixmap back;

int mixer_dev;

int Volume;
int Exit_wanted;

/* Command line parameters */
char *p_disp;

/****************************************************************
 * 
 * Configuration variables
 *
 ***************************************************************/
char *cfg_mixer;
char *cfg_foreground;
char *cfg_background;
Font cfg_font;
int cfg_timeout;
int cfg_geom_x;
int cfg_geom_y;
int cfg_geom_w;
int cfg_geom_h;
KeySym cfg_volup;
KeySym cfg_voldown;
char *cfg_volstring;


/****************************************************************
 * 
 * Draw Volume
 *
 ***************************************************************/
void draw_volume(void)
{
    XSegment pts[100];
    int x, i = 0;
    float y, yd;
    int border = cfg_geom_h / 10;
    XTextItem ti;
    char buf[100];

    if (ioctl(mixer_dev, SOUND_MIXER_READ_VOLUME, &Volume) < 0) {
	fputs("Error reading volume\n", stderr);
	Volume = 0x2020;
    }
    Volume &= 0xFF; /* Take right channel */

    y = cfg_geom_h - (border<<1);
    yd = y / 100;

    while (i < Volume) {
	pts[i].x1 = (x = border + (i * ((cfg_geom_w - (border<<1))/100.0)));
	pts[i].y1 = border + y;
	y -= yd;
	pts[i].x2 = x;
	pts[i].y2 = (cfg_geom_h - border);
	i++;
    }
    
    if (cfg_background)
	XClearWindow(display, mainwindow);
    else
	XCopyArea(display, back, mainwindow, gcon, 0, 0, cfg_geom_w, cfg_geom_h, 0, 0);

    /* Draw the volume bar */
    XDrawSegments(display, mainwindow, gcon, pts, Volume);

    snprintf(buf, sizeof(buf), "%s %u %%", cfg_volstring, Volume);
    ti.chars = buf;
    ti.nchars = strlen(buf);
    ti.delta = 10;
    ti.font = None;
    XDrawText(display, mainwindow, gcon, 50, 50, &ti, 1);
    
    // Send the "DrawLine" request to the server
    XFlush(display);
}

/****************************************************************
 * 
 * Create Main Window
 *
 ***************************************************************/
Window create_main_window(Display *dpy)
{
    XColor xcol;
    XSetWindowAttributes attr;
    Window w;
    int mask;

    /* Setup window attributes */
    attr.save_under = True; /* We are a popup window */
    attr.override_redirect = True;
    /* We are notified of structure changes (mapping of windows) and key
     * presses */
    attr.event_mask = StructureNotifyMask | KeyPressMask;
    mask = CWSaveUnder | CWOverrideRedirect | CWEventMask;

    /* Allocate a background color for the window */
    if (cfg_background) {
	XAllocNamedColor(display, DefaultColormap(display, 0), cfg_background, &xcol, &xcol);
	attr.background_pixel = xcol.pixel;
	mask |= CWBackPixel;
    }

    /* Create the window */
    w = XCreateWindow(dpy, DefaultRootWindow(dpy), cfg_geom_x, cfg_geom_y, cfg_geom_w, cfg_geom_h, 0,
	CopyFromParent, InputOutput, CopyFromParent, mask, &attr);

    // "Map" the window (that is, make it appear on the screen)
    XMapWindow(dpy, w);

    // Create a "Graphics Context"
    gcon = XCreateGC(dpy, w, 0, NULL);

    /* Allocate a fore/background colors in the colormap, and set them in the context */
    XAllocNamedColor(display, DefaultColormap(display, 0), cfg_foreground, &xcol, &xcol);
    XSetForeground(display, gcon, xcol.pixel);

    /* Set font */
    XSetFont(display, gcon, cfg_font);
    return w;
}

/****************************************************************
 * 
 * Signal Handler
 *
 ***************************************************************/
void signal_handler(int signal)
{
    if (signal == SIGALRM) {

	XUngrabKeyboard(display, CurrentTime);

	XUnmapWindow(display, mainwindow);
	XFlush(display);

	Exit_wanted = 1;
    }
}

/****************************************************************
 * 
 * Process Keypress
 *
 ***************************************************************/
void process_keypress(XKeyEvent *e, Display *dpy)
{
    int vol;
    KeySym sym = XLookupKeysym(e, 0);

    /* Direct volume setting */
    if (sym >= XK_1 && sym <= XK_9) {
	Volume = (sym - XK_0) * 10;

    /* Volume up */
    } else if (sym == cfg_volup) {
	if (Volume < 96) {
	    Volume += 3;
	} else {
	    XBell(display, 100);
	}

    /* Volume down */
    } else if (sym == cfg_voldown) {
	if (Volume > 3) {
	    Volume -= 3;
	} else {
	    XBell(display, 100);
	}
    
    /* Default case */
    } else {
	XUngrabKeyboard(dpy, CurrentTime);
	XUnmapWindow(display, mainwindow);
	Exit_wanted = 1;
	return;
    }

    /* Actually set the volume */
    vol = (Volume << 8) | Volume;
    ioctl(mixer_dev, SOUND_MIXER_WRITE_VOLUME, &vol);
    draw_volume();
    alarm(cfg_timeout);
}

/****************************************************************
 * 
 * Read Resources
 *
 ***************************************************************/
void read_resources(void)
{
    XrmDatabase db;
    char *retval;
    char *ptr;
    XrmValue xval;

    XrmInitialize();

    /* Get the resource string, either from file or from X server */
#ifdef XRESOURCES_FILE
    if (NULL == (db = XrmGetFileDatabase("/home/incore/.Xdefaults"))) {
#else
    if (NULL == (db = XrmGetStringDatabase(XResourceManagerString(display)))) {
#endif
	fputs("Error creating resource database\n", stderr);
	return;
    }

#define SAVE_RESOURCE(name, variable, defval) \
    if (True == XrmGetResource(db, "xvol." name, "Xvol", &retval, &xval)) \
    { variable = malloc(xval.size); strcpy(variable, xval.addr); } \
    else { variable = malloc(strlen(defval) + 1); strcpy(variable, defval); }

    if (!cfg_mixer) { SAVE_RESOURCE("mixer", cfg_mixer, "/dev/mixer"); }
    if (!cfg_volstring) { SAVE_RESOURCE("volstring", cfg_volstring, "Volume:"); }
    if (!cfg_foreground) { SAVE_RESOURCE("foreground", cfg_foreground, "green"); }

    /* Background handling */
    if (!cfg_background) {
	if (True == XrmGetResource(db, "xvol.background", "Xvol", &retval, &xval)) {
	    cfg_background = malloc(xval.size);
	    strcpy(cfg_background, xval.addr);
	} else {
	    cfg_background = 0;
	}
    }

    /* Here we need to convert to an integers */
    if (0 == cfg_timeout) {
	SAVE_RESOURCE("timeout", ptr, "3");
	cfg_timeout = atoi(ptr);
	cfg_timeout = (cfg_timeout > 0 ? cfg_timeout : 3);
	free(ptr);
    }

    if (!cfg_geom_x) {
	SAVE_RESOURCE("geometry", ptr, "600x100+100+100"); 
	sscanf(ptr, "%ux%u%i%i", &cfg_geom_w, &cfg_geom_h, &cfg_geom_x, &cfg_geom_y);
	cfg_geom_x = (cfg_geom_x > 10 ? cfg_geom_x : 600);
	cfg_geom_y = (cfg_geom_y > 10 ? cfg_geom_y : 100);
	cfg_geom_w = (cfg_geom_w > 10 ? cfg_geom_w : 100);
	cfg_geom_h = (cfg_geom_h > 10 ? cfg_geom_h : 100);
	free(ptr);
    }

    if (!cfg_volup) {
	SAVE_RESOURCE("volumeup", ptr, "KP_Add");
	cfg_volup = XStringToKeysym(ptr);
	free(ptr);
    }
    
    if (!cfg_voldown) {
	SAVE_RESOURCE("volumedown", ptr, "KP_Subtract");
	cfg_voldown = XStringToKeysym(ptr);
	free(ptr);
    }

    if (!cfg_font) {
	SAVE_RESOURCE("font", ptr, "-misc-fixed-medium-r-*-*-18-*-*-*-*-*-*-*");
	cfg_font = XLoadFont(display, ptr);
	free(ptr);
    }
    
    XrmDestroyDatabase(db);
}

/****************************************************************
 * 
 * Parse command line
 *
 ***************************************************************/
int parse_cmdline(int count, char **argv)
{
    int i = 1;

#define SAVE_PARM(varname, ptr) if (varname) free(varname); \
    varname = malloc(strlen(ptr)+1) ; strcpy(varname, ptr)

    argv++;
    while (i++ < count) {
	if (!strcmp("-display", *argv)) {
	    argv++; i++;
	    p_disp = *argv;

	} else if (!strcmp("-mixer", *argv)) {
	    argv++; i++;
	    SAVE_PARM(cfg_mixer, *argv);

	} else if (!strcmp("-volstring", *argv)) {
	    argv++; i++;
	    SAVE_PARM(cfg_volstring, *argv);

	} else if (!strcmp("-fg", *argv)) {
	    argv++; i++;
	    SAVE_PARM(cfg_foreground, *argv);

	} else if (!strcmp("-bg", *argv)) {
	    argv++; i++;
	    SAVE_PARM(cfg_background, *argv);
	    
	} else if (!strcmp("-timeout", *argv)) {
	    argv++; i++;
	    cfg_timeout = atoi(*argv);
	    if (cfg_timeout == 0) {
		fprintf(stderr, "Invalid timeout value: %s\n", *argv);
		return 1;
	    }

	} else if (!strcmp("-geometry", *argv)) {
	    argv++; i++;
	    if (4 != (sscanf(*argv, "%ux%u%i%i", &cfg_geom_w, &cfg_geom_h,
			&cfg_geom_x, &cfg_geom_y))) {
		fprintf(stderr, "Invalid geometry: %s\n", *argv); 
		return 1;
	    }
	    
	} else if (!strcmp("-volumeup", *argv)) {
	    argv++; i++;
	    cfg_volup = XStringToKeysym(*argv);

	} else if (!strcmp("-volumedown", *argv)) {
	    argv++; i++;
	    cfg_voldown = XStringToKeysym(*argv);

	} else if (!strcmp("-font", *argv)) {
	    argv++; i++;
	    cfg_font = XLoadFont(display, *argv);

	} else {
	    fprintf(stderr, "Unknown parameter: %s\n", *argv);
	    fputs("Parameters:\n"
		"   -display <display>    (defaults to current)\n"
		"   -mixer <device>       (defaults to '/dev/mixer')\n"
		"   -volstring <string>   (defaults to 'Volume:')\n"
		"   -fg <color>           (defaults to 'green')\n"
		"   -bg <color>           (transparent if not specified)\n"
		"   -timeout <sec>        (defaults to 3 seconds)\n"
		"   -geometry <geometry>  (defaults to '600x100+100+100')\n"
		"   -volumeup <keyname>   (defaults to 'KP_Add')\n"
		"   -volumedown <keyname> (defaults to 'KP_Subtract')\n"
		"   -font <fontname>      (defaults to 'fixed')\n"
		,stderr);
	    return 1;
	}
	argv++;
    }
    return 0;
}

/****************************************************************
 * 
 * Main
 *
 ***************************************************************/
int main(int argc, char **argv)
{
    XEvent e;
    Window lastwindow;
    int lastfocus;

    if (parse_cmdline(argc, argv))
	return 1;
    
    /* Sets handler of alarm signal */
    if (SIG_ERR == signal(SIGALRM, &signal_handler) ) {   
	fputs("Unable to set signal handler\n", stderr);
	return 1;
    }
    
    // Open the display
    if (0 == (display = XOpenDisplay(p_disp))) {
	fputs("Error opening display\n", stderr);
	return 1;
    }

    read_resources();

    if ((mixer_dev = open(cfg_mixer, O_RDONLY)) < 0) {
	fputs("Error opening mixer device\n", stderr);
	return 1;
    }
    
    mainwindow = create_main_window(display);

    /*******************/
    /* Main Event Loop */
    /*******************/
    while (!Exit_wanted) {
	XNextEvent(display, &e);
	switch (e.type) {
	case MapNotify: /* Window has just been mapped, draw something on it */

	    /* Grab background image of window */
	    back = XCreatePixmap(display, mainwindow, cfg_geom_w, cfg_geom_h, DefaultDepth(display, DefaultScreen(display)));
	    XCopyArea(display, mainwindow, back, gcon, 0, 0, cfg_geom_w, cfg_geom_h, 0, 0);
	    
	    /* Draw current volume status and grab keyboard */
	    draw_volume();
	    XGrabKeyboard(display, mainwindow, True, GrabModeAsync, GrabModeAsync, CurrentTime);
	    XGetInputFocus(display, &lastwindow, &lastfocus);
	    XSetInputFocus(display, mainwindow, RevertToParent, CurrentTime);
	    alarm(cfg_timeout);
	    break;

	case KeyPress:
	    process_keypress((XKeyEvent *) &e, display);
	    break;

	default:
	    break;
	}
    }
    
    XFreePixmap(display, back);
    XUnloadFont(display, cfg_font);
    XDestroyWindow(display, mainwindow);
    XSetInputFocus(display, lastwindow, lastfocus, CurrentTime);
    XCloseDisplay(display);

    /* Free some pointers */
    free(cfg_mixer);
    free(cfg_volstring);
    free(cfg_foreground);
    if (cfg_background) free(cfg_background);

    close(mixer_dev);
    return 0;
}

